
(* calculate a map that has the distance to each square in the layout
   for the robot. Use the 'eval' function to evaluate squares on the
   way (for instance, if we have to pass through a robot or near water
   we might consider this a "long" path, or if we pass through a base
   or package we might consider this a "short" path) *)

structure FloodFill =
struct
    open World
    type distmap = int Array.array

    val BIG = 0x3FFFFF00

    fun writeloc dm (_,w,h) (r,c) n =
        Array.update(dm, c + r * w, n)

    fun readloc dm (_, w,h) (r,c) =
        Array.sub(dm, c + r * w)

    val itos = Int.toString

    val robchars = Vector.fromList (List.map Char.toString (String.explode "0123456789"))
    val robcharslen = Vector.length robchars

    val distchars = Vector.fromList (List.map Char.toString (String.explode "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"))
    val distcharslen = Vector.length distchars

    fun clearmap dmap = Array.modify (fn _ => BIG) dmap

    (* update the existing array based on a given location *)
    fun updatemap dmap (board as BOARD{layout=layout as (b,w,h),...}) loc =
        let val _ = clearmap dmap

            val pq = PQueue.empty ()
            fun flood () =
                case PQueue.remlowest pq of
                    NONE => dmap
                  | SOME (loc as (r,c)) =>
                        let 
                            (* get its actual, most up-to-date distance. *)
                            val p = readloc dmap layout loc

			    fun dodir (l as (r, c)) =
				let in
				    (* only look at neighbors that are open
                                       and have priority greater than p+1: *)
				    case sub layout l of
					SPOT{space = Dot _, robots=rl} =>
					    let val d = 1 + length rl * 25
					    in if readloc dmap layout l > (p+d) then
						let in
						    (* now set these to have their new lower values ! *)
						    writeloc dmap layout l (p+d);
						    (* and insert them in queue *)
						    PQueue.insert pq (p+d) l
						end
					       else ()
					    end
				      | _ => ()
				end
                        in
			    dodir (r + 1, c);
			    dodir (r - 1, c);
			    dodir (r, c + 1);
			    dodir (r, c - 1);
                            (* and loop *)
                            flood ()
                        end
        in
            (* where we already are is dist 0 *)
            PQueue.insert pq 0 loc;
            writeloc dmap layout loc 0;
            (* go *)
            flood ()
        end

    (* Update the map based on a robot id *)
    fun updatemap_id dmap (board as BOARD{robots,...}) robid =
	let val (ROBOT{loc,...}) = RobUtil.getbot robots robid
        in updatemap dmap board loc
	end

    (* make a new map *)
    (*fun makemap eval (board as BOARD{layout=(_,w,h),...}) robid =
        updatemap (Array.array (w * h, BIG)) eval board robid*)
    fun makemap (BOARD{layout=(_,w,h),...}) = Array.array (w * h, BIG)

    fun mapstring dm (BOARD{layout=(board, width, height), robots, packages, ...}) =
      let
        val roblocs = List.map (fn (ri, World.ROBOT {loc, ...}) => (ri, loc)) 
	             (World.Map.listItemsi robots)
      in
        Array.foldli (fn (i, _, buf) =>
                      let
                        val col = i mod width
                        val row = (height - 1) - i div width
                        val spot = Array.sub (board, col + row * width)
			val dist = Array.sub(dm, col + row * width)
                      in
                        (if row = 0 orelse row = height-1 orelse col = 0 orelse col = width-1 then
                           buf
                         else
                           (case List.find (fn (ri, (r, c)) => r = row andalso c = col) roblocs of
                              SOME (ri, (r, c)) => buf ^ Vector.sub (robchars, ri mod robcharslen)
                            | NONE => buf ^ (case getspot spot of
						 Wall => "#"
					       | Water => "~"
					       | Dot true => "@"
					       | Dot false => 
							 if dist >= BIG then "X"
							 else Vector.sub (distchars, 
									  dist mod distcharslen)
						     )) ^ (if col = width-2 then "\n" else ""))
                      end) "" (board, 0, NONE)
      end

    (* return the move direction S E W N that gets me to the destination quickest *)
    fun dirto dmap (BOARD{layout=layout as (b,w,h),...}) (destr, destc) (atr, atc) =
	let
	    fun crawl (r, c) = 
		if r = atr andalso c = atc then (Pick [])
		else if r = atr andalso (c+1) = atc then Move W
		     else if r = atr andalso (c - 1) = atc then Move E
			  else if c = atc andalso (r + 1) = atr then Move S
			       else if c = atc andalso (r - 1) = atr then Move N
				    else let
					     val moves = [(1,0), (~1,0), (0,1), (0,~1)]
					     val moves = map (fn (dr,dc) => (readloc dmap layout (r + dr,
												  c + dc),
									     (dr,dc))) moves
					     val moves = ListUtil.sort (ListUtil.byfirst Int.compare) moves
(* 					     val _ = print ("crawl (" ^ itos r ^ ", " ^ itos c ^ "):\n") *)
(* 					     val _ = app (fn (v,(dr,dc)) => *)
(* 							  print ("   " ^ itos dr ^ "," ^ itos dc ^ ": " ^ itos v ^ "\n")) *)
(* 						               moves *)
					     val (_, (dr, dc)) = hd moves
					 in
					     crawl (r + dr, c + dc)
					 end
	in
	    (* can never reach it? *)
	    if (readloc dmap layout (destr, destc)) >= BIG then (Pick [])
	    else crawl (destr, destc)
	end

end
